# frozen_string_literal: true

class DistributedPress
  module V1
    module Social
      # A lazy loaded reference to a remote object that can access its
      # attributes directly.
      class Reference
        extend Forwardable

        # Public addressing doesn't lead to an activity
        #
        # @return [String]
        PUBLIC = 'https://www.w3.org/ns/activitystreams#Public'

        # @return [String]
        attr_reader :uri

        # @return [DistributedPress::V1::Social::Dereferencer]
        attr_reader :dereferencer

        # @param :uri [String]
        # @param :dereferencer [DistributedPress::V1::Social::Dereferencer]
        def initialize(uri:, dereferencer:)
          @uri = uri
          @dereferencer = dereferencer
        end

        # Fetches the remote object once
        #
        # @return [HTTParty::Response]
        def object
          @object ||= dereferencer.get(uri: uri)
        end

        # Is this a public post?
        #
        # @return [Bool]
        def public?
          [slice(*%w[audience to]).values].flatten.include? PUBLIC
        end

        # Is it unlisted?
        #
        # @return [Bool]
        def unlisted?
          [self['cc']].flatten.include? PUBLIC
        end

        # It's not publicly available
        #
        # @return [Bool]
        def private?
          !public? && !unlisted?
        end

        def inspect
          "#{self.class.name}(#{uri})"
        end

        def_delegators :object, :[], :dig, :to_h, :to_json, :slice
      end
    end
  end
end
